import '../expressions/expression.dart';
import '../grammar/grammar.dart';
import '../helper.dart' as helper;

class ConverterGenerator {
  final String converterName;

  final Grammar grammar;

  final bool isAsync;

  final String parserFilename;

  final String parserName;

  final List<String> _methods = [];

  ConverterGenerator({
    required this.converterName,
    required this.grammar,
    required this.isAsync,
    required this.parserFilename,
    required this.parserName,
  });

  static const _template = '''
// This code was generated by the tool, do not modify it manually
// https://pub.dev/packages/peg

{{imports}}

import '{{parser_filename}}';

export '{{parser_filename}}';

/// The [{{converter_name}}] converts data to [{{result_type}}] value.
///
/// If event-based parsing is required, then the corresponding parser instance
/// (with event handlers) must be passed as a constructor argument.
class {{converter_name}} with Converter<String, {{result_type}}> {
  final {{parser_name}} _parser;

  {{converter_name}}({
    {{parser_name}}? parser,
  }) : _parser = parser ?? {{parser_name}}();

  @override
  {{result_type}} convert(String input) {
    final result = parseString(_parser.parse{{start}}, input);
    return result;
  }

  {{methods}}
}

''';

  static const _templateStartChunkedConversion = '''
  @override
  Sink<String> startChunkedConversion(Sink<{{result_type}}> sink) {
    final input = parseAsync(_parser.parse{{start}}\$Async, (result) {
      try {
        final value = result.getResult();
        sink.add(value);
      } catch (error, stackTrace) {
        if (sink case final EventSink<Object?> sink) {
          sink.addError(error, stackTrace);
        } else {
          rethrow;
        }
      }
    });

    return input;
  }
''';

  String generate() {
    final values = <String, String>{};
    final start = grammar.start;
    if (start == null) {
      throw StateError('Grammar starting rule not found');
    }

    _methods.clear();
    final import = <String>[];
    if (isAsync) {
      import.add("import 'dart:async';");
    }

    import.add("import 'dart:convert';");
    final resultType = start.resultType ?? start.expression.resultType!;
    values['converter_name'] = converterName;
    values['imports'] = import.join('\n');
    values['parser_filename'] = parserFilename;
    values['parser_name'] = parserName;
    values['result_type'] = '$resultType';
    values['start'] = start.name;
    if (isAsync) {
      _generateStartChunkedConversion(
        resultType: resultType,
        start: start.name,
      );
    }

    values['methods'] = _methods.join('\n\n');
    return helper.render(_template, values);
  }

  void _generateStartChunkedConversion({
    required ResultType resultType,
    required String start,
  }) {
    final values = <String, String>{};
    values['result_type'] = '$resultType';
    values['start'] = start;
    final source = helper.render(_templateStartChunkedConversion, values);
    _methods.add(source);
  }
}
